from django import forms
from django.core.validators import FileExtensionValidator


class CSVFileField(forms.FileField):
    validators = [FileExtensionValidator(allowed_extensions=['csv'])]
    widget = forms.FileInput(attrs={'accept': ".csv"})


class CsvImportForm(forms.Form):
    upfile = CSVFileField(label='CSV file')


class UploadDataCsvGuessForm(CsvImportForm):
    delimiter = forms.CharField()

    def __init__(self, *args, **kwargs):
        self.headers = kwargs.pop("headers")
        super().__init__(*args, **kwargs)

        # Add a field for each expected header
        for header in self.headers:
            self.fields['header_' + header] = forms.CharField(label=header, widget=forms.Select())

    def clean_delimiter(self):
        delimiter = self.cleaned_data["delimiter"]
        if delimiter == '<Tab>':
            return '\t'
        elif delimiter == '<Space>':
            return ' '
        return delimiter



"""
Define the csv model base classe
"""
import csv
import io
import os

from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.db import DatabaseError, transaction
from django.utils.translation import gettext as _
from magic import Magic


class MetaFieldException(Exception):
    """
    Raised when no there are a field than it not is defined in the class
    """
    def __init__(self, message):
        Exception.__init__(self, message)


class ErrorMixin(object):
    def add_error(self, line_number, field, error):
        if hasattr(error, 'message_dict'):
            # message_dict attribute exists on ValidationError
            # when a dict is sent during error creation
            for field, message in error.message_dict.items():
                field = self.get_csv_field(field).replace('__all__', 'all fields')
                self.errors.append(
                    self.build_err_dict(line_number, field, ', '.join(message))
                )
            return

        elif hasattr(error, 'messages'):
            message = ', '.join(error.messages)
        else:
            message = str(error)
        self.errors.append(self.build_err_dict(line_number, field.replace('__all__', 'all fields'), message))

    def get_csv_field(self, field):
        for csv_field, real_field in self.mapping.items():
            if field == real_field:
                return csv_field
        return field

    @staticmethod
    def build_err_dict(line_number, field, message):
        return {
            'line': line_number,
            'field': field,
            'message': message
        }


class CsvModelMetaclass(type):
    def __new__(cls, name, bases, attrs):
        from djimporter.fields import Field
        new_class = super().__new__(cls, name, bases, attrs)
        for attr_name, attr_value in attrs.items():
            if isinstance(attr_value, Field) and not hasattr(attr_value, 'source'):
                setattr(attr_value, 'source', attr_name)
        return new_class


class CsvModel(ErrorMixin, metaclass=CsvModelMetaclass):

    def __init__(self, csvfile, context=None, delimiter=None, headers_mapping=None, log=None):
        self.file = csvfile
        self.context = context or {}
        self.Meta.context = context
        self.log = log

        self.errors = []
        self.list_tasks = []
        self.list_objs = []
        self.dict_error = {}
        self._meta = None

        self.fields = self.get_fields()
        self.extra_fields = getattr(self.Meta, 'extra_fields', [])
        self.headers_mapping = headers_mapping

        self.mapping = self.get_mapping()
        self.delimiter = delimiter if delimiter is not None else getattr(self.Meta, 'delimiter', ';')
        self.dbModel = self.Meta.dbModel
        self.post_save = hasattr(self.Meta, 'post_save')
        self.has_save = hasattr(self.Meta, 'save') and self.Meta.save
        self.not_create_model = hasattr(self.Meta, 'create_model') and not self.Meta.create_model
        self.unique_together = hasattr(self.Meta, 'unique_together')
        self.validate_unique = not hasattr(self.Meta, 'unique_together')
        self.append_mode = getattr(self.Meta, 'append_mode', False)
        self.exclude_fields = getattr(self.Meta, 'exclude_fields', None)

        assert not (self.unique_together and self.append_mode), (
            "Cannot set both 'unique_together' and 'append_mode' attributes: append mode will not work."
        )

    def get_user_visible_fields(self):
        # extra fields is used to capture the names of the columns used in the pre_save and
        # post_save methods each csvmodel. Is neccesary define in every Meta of csvmodel one list
        # of this names if is used in pre_save or post_save
        # This method get_user_visible_fields is used for validated the header of the file
        head = self.extra_fields.copy()
        for f in self.mapping.keys():
            field = self.fields[f]
            if hasattr(field, 'in_csv') and not field.in_csv:
                continue

            if f in self.context:
                continue

            head.append(f)

        return head

    def get_delimiter(self):
        return self.delimiter

    def get_fields(self):
        """
        Only get the names than exist in the field
        if not exist names enough for build the object
        when the it try validate the object to do crash
        """
        if hasattr(self, 'fields'):
            return self.fields

        attributes = {}
        dbModel = self.Meta.dbModel
        dmodel = {a.name: a for a in dbModel._meta.get_fields()}
        # Get all fields than is defined in the class
        for f in self.Meta.fields:
            if hasattr(self, f):
                field = getattr(self, f)
                # inject model on field to be able to access it
                field.csv_model = self.Meta.dbModel
                attributes[f] = field
            else:
                attributes[f] = dmodel[f]

        return attributes

    def get_mapping(self):
        """
        The mapping consist in a dictionary when the keys is
        the names of the columns and the value is the names
        of the fields of the model
        """
        match = {}
        for k, v in self.get_fields().items():
            if hasattr(v, 'match'):
                match[k] = v.match
            else:
                match[k] = k

        return match

    def get_dict_error(self):
        if self.dict_error:
            return self.dict_error

        msg = _("Header field '%s' is wrong")
        self.dict_error = {i: (msg % i) for i in self.get_user_visible_fields()}
        return self.dict_error

    def open_file(self, path):
        me = Magic(mime_encoding=True)
        enc = me.from_file(path)
        txt = open(path, encoding=enc).read()
        csv = bytes(txt, encoding='utf-8')
        return io.BytesIO(csv)

    def change_headers_mapping(self):
        reader = csv.DictReader(self.csv_file, delimiter=self.delimiter)
        fieldnames = reader.fieldnames

        if self.headers_mapping is None:
            return fieldnames

        # Change field name if there is a correspondence in headers_mapping
        new_fieldnames = [
            self.headers_mapping[file_header]
            if file_header in self.headers_mapping
            else file_header
            for file_header in fieldnames
        ]

        return new_fieldnames

    def is_valid(self):
        csv_file = self.file
        if isinstance(self.file, str):
            csv_file = self.open_file(self.file)
        self.csv_file = csv_file.read().decode('UTF-8').splitlines()
        fieldnames = self.change_headers_mapping()
        self.csv_reader = csv.DictReader(self.csv_file, delimiter=self.delimiter, fieldnames=fieldnames)
        # Skip header because we are passing fieldnames
        next(self.csv_reader)

        self.validate_header()
        if self.errors:
            return False

        for line_number, line in enumerate(self.csv_reader, start=2):
            # line is a dictionary with the fields of csv head as key
            # and values of the row as value of the dictionary
            self.process_line(line, line_number)

        self.validate_in_file()
        if self.errors:
            if self.has_save:
                # delete related objects created if there are errors
                # while processing post_save operations
                ids = [o.object.id for o in self.list_objs]
                self.dbModel.objects.filter(id__in=ids).delete()
            return False

        return True

    def validate_header(self):
        if self.errors: return False

        errors = {}
        for f in self.get_user_visible_fields():
            if f not in self.csv_reader.fieldnames:
                errors.update({f: _(self.get_dict_error()[f])})

        if errors:
            error = ValidationError(errors)
            self.add_error(1, 'header', error)
            return False
        return True

    def save(self):
        if self.errors: return self.errors
        if self.has_save: return
        if self.not_create_model: return

        lines = []
        for i in self.list_objs:
            if i.object:
                lines.append(i.object)
            else:
                continue

        try:
            with transaction.atomic():
                self.dbModel.objects.bulk_create(lines, batch_size=20)

                if not self.post_save: return
                for row in self.list_objs:
                    row.post_save()
                    if row.errors:
                        self.errors.extend(row.errors)

        except DatabaseError as e:
            self.add_error(1, "Error Database", {"Error Database": e.args})

    def process_line(self, line, line_number):
        data = {
            'line': line,
            'line_number': line_number,
            'context': self.context,
            'meta': self.Meta,
            'fields': self.fields,
            'mapping': self.mapping,
            'validate_unique': self.validate_unique,
            'append_mode': self.append_mode,
            'exclude_fields': self.exclude_fields
        }
        new_obj = ReadRow(**data)

        if new_obj.errors:
            self.errors.extend(new_obj.errors)
        if not new_obj.skip:
            self.list_objs.append(new_obj)

    def validate_in_file(self):
        # TODO(@slamora) optimize using Counter???
        # https://docs.python.org/3/library/collections.html#collections.Counter
        # this method is for check duplicates unique
        # and unique together fields in the same file
        # before save
        if not hasattr(self.Meta, 'unique_together'):
            return
        l_unique_together = {}
        for i in self.list_objs:
            # unique_together
            t = i.unique_together
            if t in l_unique_together:
                msg = "Combination of %s %s is repeated."
                msg = msg % (', '.join(self.Meta.unique_together), t)
                err = ValidationError({'unique': msg}, code='invalid')
                self.add_error(i.line_number, 'unique', err)
            else:
                l_unique_together[t] = None


class ReadRow(ErrorMixin):
    """
    This class build a object from the datas to a row
    """

    def __init__(self, fields=None, mapping=None, meta=None, context=None,
                 line=None, line_number=None, validate_unique=True, append_mode=False,
                 exclude_fields=None):
        self.Meta = meta
        self.fields = fields
        self.mapping = mapping
        self.context = context or {}
        self.line = line
        self.line_number = line_number
        self.validate_unique = validate_unique
        self.append_mode = append_mode
        self.exclude_fields = exclude_fields

        self.data = None
        self.object = None
        self.skip = False
        self.errors = []

        self.secuence()

    def secuence(self):
        self.get_unique_together()
        try:
            self.build_obj()
            self.create_model()
            self.pre_save()
            self.validate()
        except ValidationError:
            # stop processing the row if there are errors
            # NOTE: errors should be handled inside the functions
            # because there they have more details
            return

        if hasattr(self.Meta, 'save') and self.Meta.save:
            self.save()
            self.post_save()

    def not_create_model(self):
        return hasattr(self.Meta, 'create_model') and \
                not self.Meta.create_model

    def build_obj(self):
        data = {}
        if not self.line: return
        if self.not_create_model(): return
        for csv_fieldname, field in self.fields.items():
            model_fieldname = self.mapping[csv_fieldname]
            try:
                if hasattr(field, 'in_csv') and not field.in_csv:
                    data[model_fieldname] = field.to_python()
                    continue

                if csv_fieldname in self.context:
                    data[model_fieldname] = self.context[csv_fieldname]
                    continue

                cell = self.line[csv_fieldname]
                data[model_fieldname] = field.to_python(cell)
            except ValidationError as error:
                # handle the error here because we know which is the
                # invalid field and we want to provide this info to
                # the user.
                self.add_error(self.line_number, csv_fieldname, error)
                raise

        self.data = data

    def create_model(self):
        if not self.data: return
        self.object = self.Meta.dbModel(**self.data)

    def validate(self):
        if not self.object: return
        try:
            self.object.clean_fields(exclude=self.exclude_fields)
            self.object.clean()
            self.unique_validation()
        except ValidationError as e:
            field = list(e.message_dict.keys())[0]
            # Only print errors if field is related to uploaded file,
            # but if no errors added, add error to prevent a valid file when it isn't
            if field in list(self.mapping.values()) or len(self.errors) == 0:
                self.add_error(self.line_number, field, e)

    def unique_validation(self):
        if self.validate_unique:
            try:
                self.object.validate_unique()
            except ValidationError as e:
                if self.append_mode:
                    self.skip = True
                else:
                    raise e

    def save(self):
        if self.errors: return self.errors
        if not self.object: return
        if hasattr(self.Meta, 'create_model') and not self.Meta.create_model:
            return

        self.object.save()

    def get_unique_together(self):
        if not self.line: return
        if not hasattr(self.Meta, 'unique_together'):
            return

        t = []
        for u in self.Meta.unique_together:
            t.append(self.line[u])
        self.unique_together = tuple(t)

    def exec_f(self, f):
        try:
            f(self)
        # TODO(@slamora) check which exceptions should be caught
        #   I believe that only none or only ValidationError
        #   handling other execeptions will mask code problems
        #   e.g. missing required context
        except (ValidationError, ValueError, KeyError, ObjectDoesNotExist, DatabaseError) as error:
            file_name = os.path.split(f.__name__)[-1]
            self.add_error(self.line_number, file_name, error)

    def pre_save(self):
        if not hasattr(self.Meta, 'pre_save'): return
        for pre in self.Meta.pre_save:
            f = getattr(self.Meta, pre)
            self.exec_f(f)

    def post_save(self):
        if not hasattr(self.Meta, 'post_save'): return
        for post in self.Meta.post_save:
            f = getattr(self.Meta, post)
            self.exec_f(f)



import os
import sys

import django
from django.core import management


def pytest_addoption(parser):
    parser.addoption('--no-pkgroot', action='store_true', default=False,
                     help='Remove package root directory from sys.path, ensuring that '
                          'djimporter is imported from the installed site-packages. '
                          'Used for testing the distribution.')
    parser.addoption('--staticfiles', action='store_true', default=False,
                     help='Run tests with static files collection, using manifest '
                          'staticfiles storage. Used for testing the distribution.')


def pytest_configure(config):
    from django.conf import settings

    settings.configure(
        DEBUG_PROPAGATE_EXCEPTIONS=True,
        DATABASES={
            'default': {
                'ENGINE': 'django.db.backends.sqlite3',
                'NAME': ':memory:'
            }
        },
        SITE_ID=1,
        SECRET_KEY='not very secret in tests',
        USE_I18N=True,
        USE_L10N=True,
        STATIC_URL='/static/',
        ROOT_URLCONF='tests.urls',
        TEMPLATES=[
            {
                'BACKEND': 'django.template.backends.django.DjangoTemplates',
                'APP_DIRS': True,
                'OPTIONS': {
                    "debug": True,  # We want template errors to raise
                }
            },
        ],
        MIDDLEWARE=(
            'django.middleware.common.CommonMiddleware',
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
            'django.contrib.messages.middleware.MessageMiddleware',
        ),
        INSTALLED_APPS=(
            'django.contrib.admin',
            'django.contrib.auth',
            'django.contrib.contenttypes',
            'django.contrib.sessions',
            'django.contrib.sites',
            'django.contrib.staticfiles',
            'background_task',
            'djimporter',
            'tests',
        ),
        PASSWORD_HASHERS=(
            'django.contrib.auth.hashers.MD5PasswordHasher',
        ),
    )

    if config.getoption('--no-pkgroot'):
        sys.path.pop(0)

        # import djimporter before pytest re-adds the package root directory.
        import djimporter
        package_dir = os.path.join(os.getcwd(), 'djimporter')
        assert not djimporter.__file__.startswith(package_dir)

    # Manifest storage will raise an exception if static files are not present (ie, a packaging failure).
    if config.getoption('--staticfiles'):
        import djimporter
        settings.STATIC_ROOT = os.path.join(os.path.dirname(djimporter.__file__), 'static-root')
        settings.STATICFILES_STORAGE = 'django.contrib.staticfiles.storage.ManifestStaticFilesStorage'

    django.setup()

    if config.getoption('--staticfiles'):
        management.call_command('collectstatic', verbosity=0, interactive=False)



from datetime import datetime

from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.db.models import Manager
from django.db.models import Model as djangoModel
from django.db.models import TimeField as django_TimeField
from django.db.models.query import QuerySet
from django.utils.translation import gettext_lazy as _


class FieldError(ValueError):
    pass


class ForeignKeyFieldError(FieldError):
    def __init__(self, msg, model, value):
        self.model = model
        self.value = value
        self.msg = msg
        super(ForeignKeyFieldError, self).__init__(self.msg)


class FieldValueMissing(FieldError):
    def __init__(self, field_name):
        super(FieldValueMissing, self).__init__("No value found for field %s" % field_name)


class ListGeoException(Exception):
    """
    Raised when no there are a field than it not is defined in the class
    """
    def __init__(self, message):
        Exception.__init__(self, message)


MISSING_ERROR_MESSAGE = (
    'ValidationError raised by `{class_name}`, but error key `{key}` does '
    'not exist in the `error_messages` dictionary.'
)


class Field:
    default_error_messages = {
        'required': _('This field is required.'),
    }
    position = 0
    field_name = "Field"

    def __init__(self, *args, null=False, **kwargs):
        self.in_csv = kwargs.pop('in_csv', True)

        if 'row_num' in kwargs:
            self.position = kwargs.pop('row_num')
        else:
            self.position = Field.position
            Field.position += 1
        if 'match' in kwargs:
            self.match = kwargs.pop('match')

        if 'default' in kwargs:
            # with this value we can overwrite all values in csv
            # for this field. It is usefull when we can a default value
            # but we don't put one default value in the model
            self.has_default = self.to_python(kwargs.pop('default'))

        # If null value is allowed, this field could be empty in the row
        self.null = null

        # Collect default error message from self and parent classes
        error_messages = kwargs.get('error_messages')
        messages = {}
        for cls in reversed(self.__class__.__mro__):
            messages.update(getattr(cls, 'default_error_messages', {}))
        messages.update(error_messages or {})
        self.error_messages = messages

    def fail(self, key, **kwargs):
        """
        A helper method that simply raises a validation error.
        """
        try:
            msg = self.error_messages[key]
        except KeyError:
            class_name = self.__class__.__name__
            msg = MISSING_ERROR_MESSAGE.format(class_name=class_name, key=key)
            raise AssertionError(msg)
        message_string = msg.format(**kwargs)
        raise ValidationError(message_string, code=key)


class IntegerField(Field):
    default_error_messages = {
        'invalid': _('A valid integer is required.'),
    }
    field_name = "Integer"

    def to_python(self, value):
        if hasattr(self, "null") and not value:
            return None
        try:
            data = int(value)
        except (TypeError, ValueError):
            self.fail('invalid')
        return data


class TimeField(Field):
    field_name = "Time"

    def to_python(self, value):
        if self.null and not value:
            return None
        field = django_TimeField()
        return field.to_python(value)


class BooleanField(Field):
    field_name = "Boolean"

    def default_is_true_method(self, value):
        if hasattr(self, "null") and not value:
            return None
        return value.lower() == "true"

    def __init__(self, *args, **kwargs):
        if 'is_true' in kwargs:
            self.is_true_method = kwargs.pop('is_true')
        else:
            self.is_true_method = self.default_is_true_method
        super(BooleanField, self).__init__(*args, **kwargs)

    def to_python(self, value):
        return self.is_true_method(value)


class CharField(Field):
    field_name = "String"

    def to_python(self, value):
        if value:
            return value.strip()
        else:
            return value


class FloatField(Field):
    default_error_messages = {
        'invalid': _('A valid number is required.'),
    }
    field_name = "Float"

    def to_python(self, value):
        # handle empty values depending of this field is nullable
        if not value:
            if self.null:
                return None
            self.fail('required')
        try:
            return float(value)
        except (TypeError, ValueError):
            self.fail('invalid')


class DateField(Field):
    DEFAULT_DATE_FORMAT = '%Y-%m-%d'
    field_name = "Date"

    def __init__(self, *args, **kwargs):
        self.strptime = kwargs.pop('strptime', self.DEFAULT_DATE_FORMAT)

    def to_python(self, value):
        if value:
            try:
                return datetime.strptime(value, self.strptime)
            except ValueError as e:
                raise ValidationError(e)
        return None


class ForeignKey(Field):
    field_name = "ForeignKey"

    def __init__(self, *args, **kwargs):
        self.pk = kwargs.pop('pk', 'pk')
        if len(args) < 1:
            raise ValueError("You should provide a Model as the first argument.")
        self.model = args[0]
        try:
            if not issubclass(self.model, djangoModel):
                raise TypeError("The first argument should be a django model class.")
        except TypeError:
            raise TypeError("The first argument should be a django model class.")
        super(ForeignKey, self).__init__(**kwargs)

    def to_python(self, value):
        try:
            return self.model.objects.get(**{self.pk: value})
        except ObjectDoesNotExist:
            msg = "No match found for %(model)s with value %(value)s"
            params = {'model': self.model.__name__, 'value': value}
            raise ValidationError(msg, params=params)


class SlugRelatedField(Field):
    # We use this field for match one object in one ForeignKey
    # but we need an other field that is not a tipical id
    # Practicaly is the same of ForeignKey but from other identificatior
    # not the id or pk
    field_name = "Slug_Related_Field"
    queryset = None

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        self.args = args
        self.kwargs = kwargs
        self.queryset = kwargs.pop('queryset', self.queryset)

        if 'match' in kwargs:
            self.match = kwargs['match']
        if 'slug_field' in kwargs:
            self.slug_field = kwargs['slug_field']

        assert self.queryset is not None, (
            'Relational field must provide a `queryset` argument, '
            'override `get_queryset`, or set read_only=`True`.'
        )
        self.model = self.queryset.model

    def get_queryset(self):
        queryset = self.queryset
        if isinstance(queryset, (QuerySet, Manager)):
            # Ensure queryset is re-evaluated whenever used.
            # Note that actually a `Manager` class may also be used as the
            # queryset argument. This occurs on ModelSerializer fields,
            # as it allows us to generate a more expressive 'repr' output
            # for the field.
            # Eg: 'MyRelationship(queryset=ExampleModel.objects.all())'
            queryset = queryset.all()
        return queryset

    def get(self, value):
        # wee split in this function for help to RelatedFromUniquesField
        return self.get_queryset().get(**{self.slug_field: value.strip()})

    def to_python(self, value):
        # handle empty values depending of this field is nullable
        if not value:
            if self.null:
                return None

            # handle if null is defined on model field
            model_field = self.csv_model._meta.get_field(self.source)
            if model_field.null:
                # FIXME this is hacky, any other idea to address that?
                # monkey patch to avoid csvimporter.validate() raise ValidationError
                # this field cannot be blank while runnin `object.clean_fields()`
                model_field.blank = True
                return None

            self.fail('required')

        try:
            return self.get(value)
        except ObjectDoesNotExist:
            msg = "No match found for %(model)s with value %(value)s"
            params = {'model': self.model.__name__, 'value': value}
            raise ValidationError(msg, params=params)

        except (TypeError, ValueError) as e:
            raise ValidationError(e, code='invalid')


class CachedSlugRelatedField(Field):
    """
    SlugRelatedField which caches queryset on memory to boost importer speed.

    It performs a single "big" database query instead of N "small" queries
    where N is the number of rows to be imported.
    """
    def __init__(self, *args, null=False, **kwargs):
        queryset = kwargs.pop('queryset')
        slug_field = kwargs.pop('slug_field')

        self.queryset = queryset
        self.slug_field = slug_field
        self.model = queryset.model

        super().__init__(*args, null=null, **kwargs)

    def to_python(self, value):
        if not hasattr(self, 'cached_queryset'):
            # NOTE: cast to str dict key because CSV value by default its a string
            self.cached_queryset = {
                str(getattr(obj, self.slug_field)): obj for obj in self.queryset
            }

        value = value.strip()
        try:
            return self.cached_queryset[value]
        except KeyError:
            msg = "No match found for '%(model)s' with value '%(value)s' on field '%(slug)s'"
            params = {'model': self.model.__name__, 'value': value, 'slug': self.slug_field}
            raise ValidationError(msg, params=params, code='invalid')


class RelatedFromUniquesField(SlugRelatedField):
    # overwrite SlugRelatedField for du a query more complex
    # from fields of unique together

    field_name = "Related_Uniques_together_field"

    def get(self, dvalue):
        d = {k: dvalue[self.slug_field[k]].strip() for k in self.slug_field}
        return self.get_queryset().get(**d)


class CsvRelated(Field):
    field_name = "Csv_Related"

    def __init__(self, *args, **kwargs):
        self.csvModel = args[0]

    def to_python(self, value):
        return value


class DefaultField(Field):
    """
    We use this field for override all values in csv or add one
    that not exist in the csv

    """

    field_name = "DefaultField"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.value = args[0]

    def to_python(self):
        return self.value


class ManyToManyField(Field):
    """
    We use this field for save one Information int csv in a
    additional table that to be a realtion many to many

    """

    field_name = "ManyToManyField"

    def __init__(self, *args, **kwargs):
        self.csvModel = args[0]
        self.method = self.csvModel.__name__.lower()
        if 'dict_params' in kwargs:
            self.dict_params = kwargs['dict_params']
        if 'match' in kwargs:
            self.match = kwargs['match']

    def task(self, object_father):
        if hasattr(object_father, self.method):
            method = getattr(object_father, self.method)
            getattr(method, 'add')(self.parameters)

    def to_python(self, colname, line_number, value):
        self.colname = colname
        self.line_number = line_number
        self.dict_params[self.match] = value

        p1 = self.csvModel.objects.get(**self.dict_params)
        self.parameters = p1
        return self


class CsvFieldLink(Field):
    """
    We use this field for save the same Information in to fields
    in the model but there are only one in the csv is offer.
    As args you get a dictionary composed for k the value in the
    model and v the column name from you want copy
    normaly this key not exist in the csv file.
    And is necessary that the name of the CsvFieldLink is the same
    than the key of this dictionary

    """

    field_name = "Csv_Field_Link"

    def __init__(self, *args, **kwargs):
        self.link = args[0]

    def to_python(self, value):
        return value


class ComposedKeyField(ForeignKey):
    def to_python(self, value):
        try:
            return self.model.objects.get(**value)
        except ObjectDoesNotExist:
            raise ForeignKeyFieldError("No match found for %s" % self.model.__name__, self.model.__name__, value)


class MultiSlugRelatedField(SlugRelatedField):
    """
    overwrite SlugRelatedField for do a query more complex
    from observation of pecbms.
    The target is get one value from more than one field of them.
    Is tipicaly when the foreingkey depend of more than one field.
    """

    field_name = "MultiSlugRelatedField"

    def __init__(self, *args, **kwargs):
        super(MultiSlugRelatedField, self).__init__(*args, **kwargs)
        # matchs is similar to match but is a dictionary
        # becouse is multiple
        self.matchs = kwargs.pop('matchs', None)

    def get(self, dvalue):
        return self.get_queryset().get(**dvalue)

    def to_python(self, line):
        try:
            dvalues = {self.matchs[k]: line[k].strip() for k in self.matchs.keys()}
            return self.get(dvalues)
        except ObjectDoesNotExist:
            raise ForeignKeyFieldError("No match found for %s" % self.model.__name__, self.model.__name__, line)
        except (TypeError, ValueError, KeyError):
            raise FieldError('invalid')



"""
This tests is a collection of test for check the model and views logs
"""
from django.test import TestCase
from django.urls import reverse

from djimporter.models import ImportLog


class TestLogs(TestCase):

    def setUp(self):
        self.log_id = 1
        data = {'id': self.log_id, 'status': ImportLog.CREATED, 'user': "user1",
                'input_file': "File_one.csv"}
        self.log = ImportLog.objects.create(**data)

    def test_entrylogs_list(self):
        url = reverse('djimporter:importlog-list')
        res = self.client.get(url)
        self.assertEqual(res.status_code, 200)

    def test_entrylog_id(self):
        url = reverse('djimporter:importlog-detail', args=[self.log_id])
        res = self.client.get(url)
        self.assertEqual(res.status_code, 200)

    def test_entrylog_delete_get(self):
        url = reverse('djimporter:importlog-delete', args=[self.log_id])
        res = self.client.get(url)
        self.assertEqual(res.status_code, 200)

    def test_entrylog_delete_post(self):
        self.assertTrue(ImportLog.objects.filter(id=self.log_id).exists())
        url = reverse('djimporter:importlog-delete', args=[self.log_id])
        res = self.client.post(url)
        self.assertEqual(res.status_code, 302)
        self.assertFalse(ImportLog.objects.filter(id=self.log_id).exists())



from django.apps import apps
from django.conf import settings


def check_importlog_model(app_configs=None, **kwargs):
    if app_configs is None:
        importlog_class = getattr(settings, 'IMPORT_LOG_MODEL', 'djimporter.ImportLog')
        apps.get_model(importlog_class)
    else:
        app_label, model_name = settings.IMPORT_LOG_MODEL.split('.')
        for app_config in app_configs:
            if app_config.label == app_label:
                app_config.get_model(model_name)
                break
        else:
            # Checks might be run against a set of app configs that don't
            # include the specified user model. In this case we simply don't
            # perform the checks defined below.
            return []

    return []



# Generated by Django 3.2.17 on 2023-05-15 13:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djimporter', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='importlog',
            name='user',
            field=models.CharField(max_length=512),
        ),
    ]



from django.apps import AppConfig
from django.core import checks

from .checks import check_importlog_model


class ImporterConfig(AppConfig):
    name = 'djimporter'

    def ready(self):
        checks.register(check_importlog_model, checks.Tags.models)



from django.urls import path

from . import views

app_name = 'djimporter'

urlpatterns = [
    path('logs/', views.ListImportsView.as_view(), name='importlog-list'),
    path('logs/<int:pk>/', views.ImportDetailView.as_view(), name='importlog-detail'),
    path('logs/<int:pk>/delete/', views.ImportDeleteView.as_view(), name='importlog-delete'),
    path('logs/<int:pk>/get/', views.ImportLogGetView.as_view(), name='importlog-get')
]



"""
The `CsvModel` class is essentially shortcuts for automatically creating
serializers based on a given model class.
These tests deal with ensuring that we correctly map the model fields onto
an appropriate set of serializer fields for each case.
"""
import os

from django.core.exceptions import ValidationError
from django.test import TestCase

from djimporter import fields, importers

from .models import Album, ForeignKeySource, ForeignKeyTarget, Musician, Song

BASE_DIR = os.path.dirname(os.path.abspath(__file__))

TESTDATA_DIR = os.path.join(BASE_DIR, 'data/')


class SlugFieldMapping(TestCase):
    def test_valid(self):
        Musician.objects.bulk_create([
            Musician(name="Susan Schmith", instrument="guitar"),
            Musician(name="Johan Wolf", instrument="piano"),
        ])

        class AlbumCsv(importers.CsvModel):
            artist = fields.SlugRelatedField(slug_field="name", queryset=Musician.objects.all())

            class Meta:
                delimiter = ';'
                dbModel = Album
                fields = ['name', 'release_date', 'num_stars', 'artist']

        csv_path = os.path.join(TESTDATA_DIR, 'albums.csv')
        importer = AlbumCsv(csv_path)

        self.assertTrue(importer.is_valid())

    def test_missing_slug_related(self):
        class ForeignKeySourceCsv(importers.CsvModel):
            target = fields.SlugRelatedField(
                queryset=ForeignKeyTarget.objects.all(), slug_field='name')

            class Meta:
                dbModel = ForeignKeySource
                fields = ('name', 'target')

        ForeignKeyTarget.objects.create(name='bar')

        csv_path = os.path.join(TESTDATA_DIR, 'ForeignKeySource.csv')
        importer = ForeignKeySourceCsv(csv_path)

        self.assertFalse(importer.is_valid())


class NullableSlugFieldTest(TestCase):
    def test_valid(self):
        artist = Musician.objects.create(name="Lola", instrument="guitar")
        Album.objects.create(name="Lolailo", release_date="2023-02-02", num_stars=4, artist=artist)
        class SongCsv(importers.CsvModel):
            album = fields.SlugRelatedField(
                queryset=Album.objects.all(),
                slug_field='name',
            )

            class Meta:
                dbModel = Song
                fields = ('name', 'album')


        csv_path = os.path.join(TESTDATA_DIR, 'songs.csv')
        importer = SongCsv(csv_path)

        self.assertTrue(importer.is_valid(), importer.errors)


class CachedSlugFieldTest(TestCase):
    def test_valid(self):
        ForeignKeyTarget.objects.bulk_create([
            ForeignKeyTarget(name='bar'),
            ForeignKeyTarget(name='bar2'),
        ])

        class ForeignKeySourceCsv(importers.CsvModel):
            target = fields.CachedSlugRelatedField(
                queryset=ForeignKeyTarget.objects.all(), slug_field='name')

            class Meta:
                dbModel = ForeignKeySource
                fields = ('name', 'target')

        csv_path = os.path.join(TESTDATA_DIR, 'ForeignKeySource_valid.csv')
        importer = ForeignKeySourceCsv(csv_path)

        self.assertTrue(importer.is_valid(), importer.errors)

    def test_invalid_missing_required_target_value(self):
        ForeignKeyTarget.objects.create(name='bar')

        class ForeignKeySourceCsv(importers.CsvModel):
            target = fields.CachedSlugRelatedField(
                queryset=ForeignKeyTarget.objects.all(), slug_field='name')

            class Meta:
                dbModel = ForeignKeySource
                fields = ('name', 'target')

        csv_path = os.path.join(TESTDATA_DIR, 'ForeignKeySource.csv')
        importer = ForeignKeySourceCsv(csv_path)

        self.assertFalse(importer.is_valid())


class DateFieldTest(TestCase):
    def test_unexpected_date_format(self):
        created = fields.DateField()
        self.assertRaises(ValidationError, created.to_python, '19/04/02 07:50')


class IntegerField(TestCase):
    def test_invalid_literal(self):
        count = fields.IntegerField()
        self.assertRaises(ValidationError, count.to_python, 'U')


class TimeFieldTest(TestCase):
    def test_allow_null_values(self):
        start_time = fields.TimeField(null=True)
        self.assertIsNone(start_time.to_python(''))

    def test_explicit_disallow_null_values(self):
        start_time = fields.TimeField(null=False)
        self.assertRaises(ValidationError, start_time.to_python, '')



# Generated by Django 3.0.4 on 2020-05-15 08:15

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ImportLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.CharField(choices=[('created', 'CREATED'), ('running', 'RUNNING'), ('failed', 'FAILED'), ('completed', 'COMPLETED')], max_length=25)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('user', models.CharField(max_length=50)),
                ('errors', models.TextField(blank=True)),
                ('num_rows', models.IntegerField(null=True)),
                ('input_file', models.CharField(max_length=100)),
            ],
            options={
                'ordering': ['-created_at'],
                'abstract': False,
            },
        ),
    ]



"""
Package metadata definition.
"""
VERSION = (0, 5, 0, 'final', 1)


default_app_config = 'djimporter.apps.ImporterConfig'


def get_version():
    "Returns a PEP 386-compliant version number from VERSION."
    if (len(VERSION) != 5 or
            VERSION[3] not in ('alpha', 'beta', 'rc', 'final')):
        raise ValueError(
            "{} is not PEP 386-compliant version number".format(VERSION))

    # Now build the two parts of the version number:
    # main = X.Y[.Z]
    # sub = .devN - for pre-alpha releases
    #     | {a|b|c}N - for alpha, beta and rc releases

    parts = 2 if VERSION[2] == 0 else 3
    main = '.'.join(str(x) for x in VERSION[:parts])

    sub = ''

    if VERSION[3] != 'final':
        mapping = {'alpha': 'a', 'beta': 'b', 'rc': 'c'}
        sub = mapping[VERSION[3]] + str(VERSION[4])

    return str(main + sub)


def get_importlog_model():
    """
    Return the ImportLog model that is active in this project.
    """
    # avoid circular dependencies (setup.py calls get_version())
    from django.apps import apps as django_apps
    from django.conf import settings
    from django.core.exceptions import ImproperlyConfigured

    importlog_class = getattr(settings, 'IMPORT_LOG_MODEL', 'djimporter.ImportLog')
    try:
        return django_apps.get_model(importlog_class, require_ready=False)
    except ValueError:
        raise ImproperlyConfigured("IMPORT_LOG_MODEL must be of the form 'app_label.model_name'")
    except LookupError:
        raise ImproperlyConfigured(
            "IMPORT_LOG_MODEL refers to model '%s' that has not been installed" % settings.IMPORT_LOG_MODEL
        )



import json

from django.db import models


class AbstractBaseLog(models.Model):
    CREATED = 'created'
    RUNNING = 'running'
    FAILED = 'failed'
    COMPLETED = 'completed'
    STATUS_CHOICES = (
        (CREATED, 'CREATED'),
        (RUNNING, 'RUNNING'),
        (FAILED, 'FAILED'),
        (COMPLETED, 'COMPLETED'),
    )

    status = models.CharField(max_length=25, choices=STATUS_CHOICES)
    created_at = models.DateTimeField(auto_now_add=True)
    user = models.CharField(max_length=512)

    errors = models.TextField(blank=True)
    num_rows = models.IntegerField(null=True)
    input_file = models.CharField(max_length=100)

    class Meta:
        abstract = True
        ordering = ['-created_at']

    def list_errors(self):
        return json.loads(self.errors or "[]")


class ImportLog(AbstractBaseLog):
    pass



import json
import os

from django.utils.module_loading import import_string
from background_task import background

from . import get_importlog_model

ImportLog = get_importlog_model()


@background(schedule=0)
def run_importer(csv_model, csv_filepath, log_id, context={}, delimiter=None, headers_mapping=None):
    """
    csv_model: should be string dotted_path e.g. 'djimporter.FooCsv'
    context: should be serializable
    """
    importer_class = import_string(csv_model)
    # mark task as running
    log = ImportLog.objects.get(id=log_id)
    log.status = ImportLog.RUNNING
    log.save()

    # run importer
    try:
        importer = importer_class(
            csv_filepath, context=context, delimiter=delimiter, headers_mapping=headers_mapping, log=log
        )
        importer.is_valid()
        importer.save()

        # update log with import result
        if importer.errors:
            log.status = ImportLog.FAILED
            log.errors = json.dumps(importer.errors)
        else:
            log.status = ImportLog.COMPLETED
            log.num_rows = len(importer.list_objs)

    except Exception as e:
        # Not controlled errors will be thrown to log
        errors = [{'line': 1, 'field': 'Internal Error', 'message': e.args}]
        log.status = ImportLog.FAILED
        log.errors = json.dumps(errors)

    log.save()

    # clean up
    os.remove(csv_filepath)



import os
from distutils.util import convert_path

from setuptools import find_packages, setup


main_ns = {}
ver_path = convert_path('djimporter/__init__.py')
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

version = main_ns['get_version']()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


setup(
    name="djimporter",
    version=version,
    url='https://github.com/ico-apps/django-importer',
    author='ICO - Institut Catala Ornitologia',
    author_email='ico@ornitologia.org',
    description='Django importer, another CSV import library.',
    long_description=('Django importer, another CSV import library.'),
    license='BSD-3-Clause',
    packages=find_packages(),
    include_package_data=True,
    install_requires=["django>=2.2,<4.0", "django-background-tasks==1.2.5", "python-magic==0.4.24"],
    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
    ],
)



from django.core.files.storage import default_storage
from django.http import HttpResponseRedirect, JsonResponse
from django.urls import reverse
from django.views import View
from django.views.generic.detail import DetailView
from django.views.generic.edit import FormView, DeleteView
from django.views.generic.list import ListView

from . import get_importlog_model
from .forms import CsvImportForm, UploadDataCsvGuessForm
from .tasks import run_importer

ImportLog = get_importlog_model()


class ListImportsView(ListView):
    model = ImportLog
    template_name = "djimporter/importlog_list.html"
    url_detail = 'djimporter:importlog-detail'
    url_delete = 'djimporter:importlog-delete'

    def get_context_data(self):
        context = super().get_context_data()
        context.update({
            'url_detail': self.url_detail,
            'url_delete': self.url_delete,
        })
        return context


class ImportDetailView(DetailView):
    model = ImportLog
    template_name = "djimporter/importlog_detail.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["importlog_status_url"] = self.get_status_url()
        return context

    def get_status_url(self):
        return reverse("djimporter:importlog-get", args=(self.object.pk,))


class ImportLogGetView(View):

    def get(self, request, *args, **kwargs):
        import_log = ImportLog.objects.get(pk=self.kwargs['pk'])
        return JsonResponse({'status': import_log.status, 'id': self.kwargs['pk']}, safe=False)


class ImportDeleteView(DeleteView):
    model = ImportLog
    template_name = "djimporter/importlog_confirm_delete.html"

    def get_success_url(self, *args, **kwargs):
        return reverse('djimporter:importlog-list')


class ImportFormView(FormView):
    form_class = CsvImportForm
    importer_class = None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        importer_class = self.get_importer_class()
        importer = importer_class('')
        context['importer'] = importer
        return context

    def form_valid(self, form):
        """
        After calling this method, valid form will be available on `self.form`
        """
        # store valid form to allow other methods access it
        # e.g. get_importer_context
        self.form = form

        # Create a dict that maps current file headers with expected headers names (fields and extra_fiels in Meta).
        # The dict has the following format:
        # {
        #   file_header_value1: importer_header_value1,
        #   file_header_value2: importer_header_value2,
        #   ...
        # }
        header_mapping = {}
        for field_name in form.cleaned_data:
            if field_name.startswith('header_'):
                header_mapping[form.cleaned_data[field_name]] = field_name.replace('header_', '', 1)

        kwargs = {}
        if 'delimiter' in form.cleaned_data:
            kwargs['delimiter'] = form.cleaned_data['delimiter']

        if len(header_mapping) > 0:
            kwargs['headers_mapping'] = header_mapping

        self.task_log = self.create_import_task(form.files['upfile'], **kwargs)

        return HttpResponseRedirect(self.get_success_url())

    def create_import_task(self, csv_file, delimiter=None, headers_mapping=None):
        importer_class = self.get_importer_class()
        task_log = self.create_import_log(csv_file)

        # save file to persistent storage
        name = default_storage.save(csv_file.name, csv_file)
        csv_path = default_storage.path(name)

        dotted_path = "{module}.{name}".format(
            module=importer_class.__module__,
            name=importer_class.__name__,
        )

        context = self.get_importer_context()
        run_importer(dotted_path, csv_path, task_log.id, context=context,
                     delimiter=delimiter, headers_mapping=headers_mapping)

        return task_log

    def create_import_log(self, csv_file):
        max_user_length = ImportLog.user.field.max_length
        user_repr = str(self.request.user)[:max_user_length]

        task_log = ImportLog.objects.create(
            status=ImportLog.CREATED,
            user=user_repr,
            input_file=csv_file.name,
        )
        return task_log

    def get_importer_class(self):
        """
        Return the class to use for the importer.
        Defaults to using `self.importer_class`.

        """
        assert self.importer_class is not None, (
            "'%s' should either include a `importer_class` attribute, "
            "or override the `get_serializer_class()` method."
            % self.__class__.__name__
        )

        return self.importer_class

    def get_importer_context(self):
        return {}

    def get_success_url(self):
        return reverse('djimporter:importlog-detail', kwargs={'pk': self.task_log.id})


class ImportFormGuessCsvView(ImportFormView):
    form_class = UploadDataCsvGuessForm

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()

        headers = self.importer_class.Meta.fields.copy()
        if hasattr(self.importer_class.Meta, 'extra_fields'):
            headers.extend(self.importer_class.Meta.extra_fields)
        kwargs.update({
            'headers': headers
        })
        return kwargs



from django.db import models


class DJImporterModel(models.Model):
    """
    Base for test models that sets app_label, so they play nicely.
    """

    class Meta:
        app_label = 'tests'
        abstract = True


class ForeignKeyTarget(DJImporterModel):
    name = models.CharField(max_length=100)


class ForeignKeySource(DJImporterModel):
    name = models.CharField(max_length=100)
    target = models.ForeignKey(ForeignKeyTarget, related_name='sources',
                               help_text='Target', verbose_name='Target',
                               on_delete=models.CASCADE)


class Musician(models.Model):
    name = models.CharField(max_length=50, unique=True)
    instrument = models.CharField(max_length=100)


class Album(models.Model):
    name = models.CharField(max_length=100)
    release_date = models.DateField()
    num_stars = models.IntegerField()
    artist = models.ForeignKey(Musician, on_delete=models.CASCADE)


class Song(models.Model):
    name = models.CharField(max_length=100)
    album = models.ForeignKey(Album, null=True, on_delete=models.SET_NULL)
